"""
Adds to Flask-SQLAlchemy the capability to atomically send
messages (signals) over a message bus.
"""

import time
import logging
from functools import wraps
from sqlalchemy import event, inspect, and_
from sqlalchemy.exc import DBAPIError

__all__ = ['SignalBus', 'SignalBusMixin']


ERROR_CODE_ATTRS = ['pgcode', 'sqlstate']
DEADLOCK_ERROR_CODES = ['40001', '40P01']
SIGNALS_TO_FLUSH_SESSION_INFO_KEY = 'flask_signalbus__signals_to_flush'
FLUSHMANY_LIMIT = 1000


class DBSerializationError(Exception):
    """The transaction is rolled back due to a race condition."""


def get_db_error_code(exception):
    """Return 5-character SQLSTATE code, or '' if not available.

    Currently `psycopg2`, `psycopg2cffi`, and `MySQL Connector` are supported.
    """

    for attr in ERROR_CODE_ATTRS:
        error_code = getattr(exception, attr, '')
        if error_code:
            break
    return error_code


def retry_on_deadlock(session, retries=6, min_wait=0.1, max_wait=10.0):
    """Return function decorator that executes the function again in case of a deadlock."""

    def decorator(action):
        """Function decorator that retries `action` in case of a deadlock."""

        @wraps(action)
        def f(*args, **kwargs):
            num_failures = 0
            while True:
                try:
                    return action(*args, **kwargs)
                except (DBAPIError, DBSerializationError) as e:
                    num_failures += 1
                    is_serialization_error = (
                        isinstance(e, DBSerializationError) or
                        get_db_error_code(e.orig) in DEADLOCK_ERROR_CODES
                    )
                    if num_failures > retries or not is_serialization_error:
                        raise
                session.rollback()
                wait_seconds = min(max_wait, min_wait * 2 ** (num_failures - 1))
                time.sleep(wait_seconds)

        return f

    return decorator


def _raise_error_if_not_signal_model(model):
    if not hasattr(model, 'send_signalbus_message'):
        raise RuntimeError(
            '{} can not be flushed because it does not have a'
            ' "send_signalbus_message" method.'
        )


class SignalBusMixin(object):
    """A **mixin class** that can be used to extend the
    `flask_sqlalchemy.SQLAlchemy` class to handle signals.

    For example::

        from flask import Flask
        from flask_sqlalchemy import SQLAlchemy
        from flask_signalbus import SignalBusMixin

        class CustomSQLAlchemy(SignalBusMixin, SQLAlchemy):
            pass

        app = Flask(__name__)
        db = CustomSQLAlchemy(app)
        db.signalbus.flush()

    """

    def init_app(self, app, *args, **kwargs):
        super(SignalBusMixin, self).init_app(app, *args, **kwargs)
        self.signalbus._init_app(app)

    @property
    def signalbus(self):
        """The associated `SignalBus` object."""

        try:
            signalbus = self.__signalbus
        except AttributeError:
            signalbus = self.__signalbus = SignalBus(self, init_app=False)
        return signalbus


class SignalBus(object):
    """Instances of this class automatically send signal messages that
    have been recorded in the SQL database, over a message
    bus. Normally, the sending of the recorded messages (if there are
    any) is done after each transaction commit, but it also can be
    triggered explicitly by a command.

    :param db: The `flask_sqlalchemy.SQLAlchemy` instance

    For example::

        from flask_sqlalchemy import SQLAlchemy
        from flask_signalbus import SignalBus

        app = Flask(__name__)
        db = SQLAlchemy(app)
        signalbus = SignalBus(db)
        signalbus.flush()

    """

    def __init__(self, db, init_app=True):
        self.db = db
        self.signal_session = self.db.create_scoped_session({'expire_on_commit': False})
        self.logger = logging.getLogger(__name__)
        self._autoflush = True
        retry = retry_on_deadlock(self.signal_session, retries=10, max_wait=1.0)
        self._flush_signals_with_retry = retry(self._flush_signals)
        event.listen(self.db.session, 'transient_to_pending', self._transient_to_pending_handler)
        event.listen(self.db.session, 'after_commit', self._safe_after_commit_handler)
        event.listen(self.db.session, 'after_rollback', self._after_rollback_handler)
        if init_app:
            if db.app is None:
                raise RuntimeError(
                    'No application found. The SQLAlchemy instance passed to'
                    ' SignalBus should be constructed with an application.'
                )
            self._init_app(db.app)

    @property
    def autoflush(self):
        """Setting this property to `False` instructs the `SignalBus` instance
        to not automatically flush pending signals after each
        transaction commit. Setting it back to `True` restores the
        default behavior.

        """

        return self._autoflush

    @autoflush.setter
    def autoflush(self, value):
        self._autoflush = bool(value)

    def get_signal_models(self):
        """Return all signal types in a list.

        :rtype: list(`signal-model`)

        """

        base = self.db.Model
        return [
            cls for cls in base._decl_class_registry.values() if (
                isinstance(cls, type) and
                issubclass(cls, base) and
                hasattr(cls, 'send_signalbus_message')
            )
        ]

    def flush(self, models=None, wait=3.0):
        """Send all pending signals over the message bus.

        :param models: If passed, flushes only signals of the specified types.
        :type models: list(`signal-model`) or `None`
        :param float wait: The number of seconds the method will wait
            after obtaining the list of pending signals, to allow
            concurrent senders to complete
        :return: The total number of signals that have been sent

        """

        models_to_flush = self.get_signal_models() if models is None else models
        pks_to_flush = {}
        try:
            for model in models_to_flush:
                _raise_error_if_not_signal_model(model)
                m = inspect(model)
                pk_attrs = [m.get_property_by_column(c).class_attribute for c in m.primary_key]
                pks_to_flush[model] = self.signal_session.query(*pk_attrs).all()
            self.signal_session.rollback()
            time.sleep(wait)
            return sum(
                self._flush_signals_with_retry(model, pk_values_set=set(pks_to_flush[model]))
                for model in models_to_flush
            )
        finally:
            self.signal_session.remove()

    def flushmany(self):
        """Send a potentially huge number of pending signals over the message bus.

        This method assumes that the number of pending signals might
        be huge, so that they might not fit into memory. However,
        `SignalBus.flushmany` is not very smart in handling concurrent
        senders. It is mostly useful when recovering from long periods
        of disconnectedness from the message bus.

        :return: The total number of signals that have been sent

        """

        models_to_flush = self.get_signal_models()
        try:
            return sum(self._flushmany_signals(model) for model in models_to_flush)
        finally:
            self.signal_session.remove()

    def _init_app(self, app):
        from . import cli

        if not hasattr(app, 'extensions'):
            app.extensions = {}
        if app.extensions.get('signalbus') not in [None, self]:
            raise RuntimeError('Can not attach more than one SignalBus to one application.')
        app.extensions['signalbus'] = self
        app.cli.add_command(cli.signalbus)

        @app.teardown_appcontext
        def shutdown_signal_session(response_or_exc):
            self.signal_session.remove()
            return response_or_exc

    def _transient_to_pending_handler(self, session, instance):
        model = type(instance)
        if hasattr(model, 'send_signalbus_message') and getattr(model, 'signalbus_autoflush', True):
            signals_to_flush = session.info.setdefault(SIGNALS_TO_FLUSH_SESSION_INFO_KEY, set())
            signals_to_flush.add(instance)

    def _after_commit_handler(self, session):
        signals_to_flush = session.info.pop(SIGNALS_TO_FLUSH_SESSION_INFO_KEY, set())
        if self.autoflush and signals_to_flush:
            signals = [self.signal_session.merge(s, load=False) for s in signals_to_flush]
            for signal in signals:
                try:
                    signal.send_signalbus_message()
                except:
                    self.logger.exception('Caught error while sending %s.', signal)
                    self.signal_session.rollback()
                    return
                self.signal_session.delete(signal)
            self.signal_session.commit()
            self.signal_session.expire_all()

    def _after_rollback_handler(self, session):
        session.info.pop(SIGNALS_TO_FLUSH_SESSION_INFO_KEY, None)

    def _safe_after_commit_handler(self, session):
        try:
            return self._after_commit_handler(session)
        except DBAPIError as e:
            if get_db_error_code(e.orig) not in DEADLOCK_ERROR_CODES:
                self.logger.exception('Caught database error during autoflush.')
            self.signal_session.rollback()

    def _get_lock_for_update(self, pk_attrs, pk_values):
        clause = and_(*[attr == value for attr, value in zip(pk_attrs, pk_values)])
        query = self.signal_session.query(*pk_attrs).filter(clause).with_for_update()
        return query.one_or_none() is not None

    def _flushmany_signals(self, model):
        self.logger.warning('Flushing %s in "flushmany" mode.', model.__name__)
        sent_count = 0
        while True:
            n = self._flush_signals(model, max_count=FLUSHMANY_LIMIT)
            sent_count += n
            if n < FLUSHMANY_LIMIT:
                break
        return sent_count

    def _flush_signals(self, model, pk_values_set=None, max_count=None):
        query = self.signal_session.query(model)
        if max_count is None:
            self.logger.info('Flushing %s.', model.__name__)
        else:
            query = query.limit(max_count)
        signals = query.all()
        self.signal_session.commit()
        burst_count = int(getattr(model, 'signalbus_burst_count', 1))
        sent_count = 0
        m = inspect(model)
        pk_attrs = [m.get_property_by_column(c).class_attribute for c in m.primary_key]
        for signal in signals:
            pk_values = m.primary_key_from_instance(signal)
            if pk_values_set is None or pk_values in pk_values_set:
                if not self._get_lock_for_update(pk_attrs, pk_values):
                    # The row has been deleted by a concurrent sender.
                    continue
                signal.send_signalbus_message()
                self.signal_session.delete(signal)
                sent_count += 1
                if sent_count % burst_count == 0:
                    self.signal_session.commit()
        self.signal_session.commit()
        self.signal_session.expire_all()
        return sent_count



import pytest
import flask_sqlalchemy as fsa
from flask_signalbus import SignalBus, retry_on_deadlock, DBSerializationError
from .conftest import SignalBusAlchemy
from sqlalchemy.exc import DBAPIError


def test_create_signalbus_alchemy(app):
    assert 'signalbus' not in app.extensions
    db = SignalBusAlchemy(app)
    assert 'signalbus' in app.extensions
    assert len(db.signalbus.get_signal_models()) == 0


def test_create_signalbus_alchemy_init_app(app):
    db = SignalBusAlchemy()
    assert 'signalbus' not in app.extensions
    db.init_app(app)
    assert 'signalbus' in app.extensions
    assert len(db.signalbus.get_signal_models()) == 0


def test_create_signalbus_directly(app):
    assert 'signalbus' not in app.extensions
    db = fsa.SQLAlchemy(app)
    signalbus = SignalBus(db)
    assert 'signalbus' in app.extensions
    assert not hasattr(db, 'signalbus')
    assert len(signalbus.get_signal_models()) == 0


def test_create_two_signalbuses_directly(app):
    db = fsa.SQLAlchemy(app)
    SignalBus(db)
    with pytest.raises(RuntimeError):
        SignalBus(db)


def test_create_signalbus_directly_no_app():
    db = fsa.SQLAlchemy()
    with pytest.raises(RuntimeError):
        SignalBus(db)


def test_flush_signal_model(app, signalbus, Signal):
    assert len(signalbus.get_signal_models()) == 1
    signalbus.flush([Signal], wait=0.0)


def test_flush_nonsignal_model(app, signalbus, NonSignal):
    assert len(signalbus.get_signal_models()) == 0
    with pytest.raises(RuntimeError):
        signalbus.flush([NonSignal], wait=0.0)


def test_flush_all_signal_models(app, signalbus, Signal, NonSignal):
    assert len(signalbus.get_signal_models()) == 1
    signalbus.flush(wait=0.0)


def test_flushmany_signal_model(app, signalbus_with_pending_signal):
    assert signalbus_with_pending_signal.flushmany() == 1


def test_send_signal_success(db, signalbus, send_mock, Signal):
    sig = Signal(name='signal', value='1')
    db.session.add(sig)
    db.session.flush()
    sig_id = sig.id
    send_mock.assert_not_called()
    db.session.commit()
    send_mock.assert_called_once_with(sig_id, 'signal', '1', {})
    assert Signal.query.count() == 0


def test_send_signal_error(db, signalbus, send_mock, Signal):
    sig = Signal(name='error', value='1')
    db.session.add(sig)
    db.session.commit()
    assert send_mock.call_count == 1
    with pytest.raises(ValueError):
        signalbus.flush(wait=0.0)
    assert send_mock.call_count == 2
    assert Signal.query.count() == 1


def test_autoflush_false(db, signalbus, send_mock, Signal):
    db.session.add(Signal(name='signal', value='1'))
    signalbus.autoflush = False
    db.session.commit()
    assert send_mock.call_count == 0
    assert Signal.query.count() == 1
    signalbus.flush(wait=0.0)
    assert send_mock.call_count == 1
    assert Signal.query.count() == 0

    db.session.add(Signal(name='signal', value='2'))
    signalbus.autoflush = True
    db.session.commit()
    assert send_mock.call_count == 2
    assert Signal.query.count() == 0


def test_model_autoflush_false(db, signalbus, send_mock, Signal):
    Signal.signalbus_autoflush = False
    db.session.add(Signal(name='signal', value='1'))
    db.session.commit()
    assert send_mock.call_count == 0
    assert Signal.query.count() == 1


def test_send_nonsignal_model(db, signalbus, send_mock, NonSignal):
    assert len(signalbus.get_signal_models()) == 0
    db.session.add(NonSignal())
    db.session.flush()
    db.session.commit()
    assert NonSignal.query.count() == 1


def test_signal_with_props_success(db, send_mock, Signal, SignalProperty):
    sig = Signal(name='signal', value='1')
    sig.properties = [
        SignalProperty(name='first_name', value='John'),
        SignalProperty(name='last_name', value='Doe'),
    ]
    db.session.add(sig)
    db.session.flush()
    assert Signal.query.count() == 1
    assert SignalProperty.query.count() == 2
    sig_id = sig.id
    send_mock.assert_not_called()
    db.session.commit()
    send_mock.assert_called_once()
    args, kwargs = send_mock.call_args
    assert kwargs == {}
    assert len(args) == 4
    assert args[:3] == (sig_id, 'signal', '1')
    props = args[3]
    assert type(props) is dict
    assert len(props) == 2
    assert props['first_name'] == 'John'
    assert props['last_name'] == 'Doe'
    assert Signal.query.count() == 0
    assert SignalProperty.query.count() == 0


def test_signal_with_props_is_efficient(app, db, Signal, SignalProperty):
    with app.test_request_context():
        sig = Signal(name='signal', value='1')
        sig.properties = [
            SignalProperty(name='first_name', value='John'),
            SignalProperty(name='last_name', value='Doe'),
        ]
        db.session.add(sig)
        db.session.commit()
        all_statements = [q.statement for q in fsa.get_debug_queries()]
        assert not any('SELECT' in s for s in all_statements)


def test_flush_signal_with_props(db, signalbus, send_mock, Signal, SignalProperty):
    signalbus.autoflush = False
    sig = Signal(name='signal', value='1')
    sig.properties = [
        SignalProperty(name='first_name', value='John'),
        SignalProperty(name='last_name', value='Doe'),
    ]
    db.session.add(sig)
    db.session.commit()
    send_mock.assert_not_called()
    signalbus.flush(wait=0.0)
    send_mock.assert_called_once()
    props = send_mock.call_args[0][3]
    assert type(props) is dict
    assert len(props) == 2
    assert props['first_name'] == 'John'
    assert props['last_name'] == 'Doe'
    assert Signal.query.count() == 0
    assert SignalProperty.query.count() == 0


def test_retry_on_deadlock(db):
    retry = retry_on_deadlock(db.session, retries=5, max_wait=0.0)
    executions = []

    @retry
    def f():
        executions.append(1)
        raise DBSerializationError

    @retry
    def g():
        executions.append(1)
        db.session.execute('xxx')

    with pytest.raises(DBSerializationError):
        f()
    assert len(executions) == 6

    with pytest.raises(DBAPIError):
        g()
    assert len(executions) == 7



import click
from flask.cli import with_appcontext
from flask import current_app
from . import SignalBus


@click.group()
def signalbus():
    """Perform SignalBus operations."""


@signalbus.command()
@with_appcontext
@click.option('-e', '--exclude', multiple=True, help='Do not flush signals with the specified name.')
@click.option('-w', '--wait', type=float, help='The number of seconds "flush" will wait after'
              ' obtaining the list of pending signals, to allow concurrent senders to complete.'
              ' The default is %s seconds.' % SignalBus.flush.__defaults__[1])
@click.argument('signal_names', nargs=-1)
def flush(signal_names, exclude, wait):
    """Send pending signals over the message bus.

    If a list of SIGNAL_NAMES is specified, flushes only those
    signals. If no SIGNAL_NAMES are specified, flushes all signals.

    """

    signalbus = current_app.extensions['signalbus']
    signal_names = set(signal_names)
    exclude = set(exclude)
    models_to_flush = signalbus.get_signal_models()
    if signal_names and exclude:
        click.echo('Warning: Specified both SIGNAL_NAMES and exclude option.')
    if signal_names:
        wrong_signal_names = signal_names - {m.__name__ for m in models_to_flush}
        models_to_flush = [m for m in models_to_flush if m.__name__ in signal_names]
    else:
        wrong_signal_names = exclude - {m.__name__ for m in models_to_flush}
    for name in wrong_signal_names:
        click.echo('Warning: A signal with name "{}" does not exist.'.format(name))
    models_to_flush = [m for m in models_to_flush if m.__name__ not in exclude]
    if wait is not None:
        signal_count = signalbus.flush(models_to_flush, wait=max(0.0, wait))
    else:
        signal_count = signalbus.flush(models_to_flush)
    if signal_count == 1:
        click.echo('{} signal has been successfully processed.'.format(signal_count))
    elif signal_count > 1:
        click.echo('{} signals have been successfully processed.'.format(signal_count))


@signalbus.command()
@with_appcontext
def flushmany():
    """Send a potentially huge number of pending signals over the message bus.

    This command assumes that the number of pending signals might be
    huge, so that they might not fit into memory. However, it is not
    very smart in handling concurrent senders. It is mostly useful
    when recovering from long periods of disconnectedness from the
    message bus.

    """

    signalbus = current_app.extensions['signalbus']
    signal_count = signalbus.flushmany()
    if signal_count == 1:
        click.echo('{} signal has been successfully processed.'.format(signal_count))
    elif signal_count > 1:
        click.echo('{} signals have been successfully processed.'.format(signal_count))


@signalbus.command()
@with_appcontext
def signals():
    """Show all signal types."""

    signalbus = current_app.extensions['signalbus']
    for signal_model in signalbus.get_signal_models():
        click.echo(signal_model.__name__)


@signalbus.command()
@with_appcontext
def pending():
    """Show the number of pending signals by signal type."""

    signalbus = current_app.extensions['signalbus']
    pending = []
    total_pending = 0
    for signal_model in signalbus.get_signal_models():
        count = signal_model.query.count()
        if count > 0:
            pending.append((count, signal_model.__name__))
        total_pending += count
    if pending:
        pending.sort()
        max_chars = len(str(pending[-1][0]))
        for n, signal_name in pending:
            click.echo('{} of type "{}"'.format(str(n).rjust(max_chars), signal_name))
    click.echo(25 * '-')
    click.echo('Total pending: {} '.format(total_pending))
